'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref8) {
  var t = _ref8.types;
  var template = _ref8.template;
  var options = _ref8.options;

  function _ref(_id) {
    if (!Plugin(_id)) {
      throw new TypeError('Function return value violates contract, expected Plugin got ' + (_id === null ? 'null' : (typeof _id === 'undefined' ? 'undefined' : _typeof(_id)) === 'object' && _id.constructor ? _id.constructor.name || '[Unknown Object]' : typeof _id === 'undefined' ? 'undefined' : _typeof(_id)));
    }

    return _id;
  }

  var PRECONDITION_NAME = 'pre';
  var POSTCONDITION_NAME = 'post';
  var INVARIANT_NAME = 'invariant';
  var ASSERT_NAME = 'assert';
  var RETURN_NAME = 'it';
  var OLD_VALUE_NAME = 'old';
  var returnId = t.identifier(RETURN_NAME);

  if (!Identifier(returnId)) {
    throw new TypeError('Value of variable "returnId" violates contract, expected Identifier got ' + (returnId === null ? 'null' : (typeof returnId === 'undefined' ? 'undefined' : _typeof(returnId)) === 'object' && returnId.constructor ? returnId.constructor.name || '[Unknown Object]' : typeof returnId === 'undefined' ? 'undefined' : _typeof(returnId)));
  }

  var guard = template('\n    if (!condition) {\n      throw new Error(message);\n    }\n  ');

  if (!(typeof guard === 'function')) {
    throw new TypeError('Value of variable "guard" violates contract, expected ({ [key: string]: Node\n}) => Node got ' + (guard === null ? 'null' : (typeof guard === 'undefined' ? 'undefined' : _typeof(guard)) === 'object' && guard.constructor ? guard.constructor.name || '[Unknown Object]' : typeof guard === 'undefined' ? 'undefined' : _typeof(guard)));
  }

  var guardFn = template('\n    const id = (it) => {\n      conditions;\n      return it;\n    }\n  ');

  if (!(typeof guardFn === 'function')) {
    throw new TypeError('Value of variable "guardFn" violates contract, expected ({ [key: string]: Node\n}) => Node got ' + (guardFn === null ? 'null' : (typeof guardFn === 'undefined' ? 'undefined' : _typeof(guardFn)) === 'object' && guardFn.constructor ? guardFn.constructor.name || '[Unknown Object]' : typeof guardFn === 'undefined' ? 'undefined' : _typeof(guardFn)));
  }

  function assemblePrecondition(path) {
    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract, expected NodePath got ' + (body === null ? 'null' : (typeof body === 'undefined' ? 'undefined' : _typeof(body)) === 'object' && body.constructor ? body.constructor.name || '[Unknown Object]' : typeof body === 'undefined' ? 'undefined' : _typeof(body)));
    }

    var fn = path.getFunctionParent();

    if (!NodePath(fn)) {
      throw new TypeError('Value of variable "fn" violates contract, expected NodePath got ' + (fn === null ? 'null' : (typeof fn === 'undefined' ? 'undefined' : _typeof(fn)) === 'object' && fn.constructor ? fn.constructor.name || '[Unknown Object]' : typeof fn === 'undefined' ? 'undefined' : _typeof(fn)));
    }

    var name = fn.node.id ? '"' + fn.node.id.name + '" ' : ' ';
    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
      }

      var message = undefined;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      } else {
        message = t.stringLiteral('Function ' + name + 'precondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      }
      path.replaceWith(guard({
        condition: staticCheck(condition),
        message: message
      }));
      return;
    }

    body.traverse({
      "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
        if (!NodePath(item)) {
          throw new TypeError('Value of argument "item" violates contract, expected NodePath got ' + (item === null ? 'null' : (typeof item === 'undefined' ? 'undefined' : _typeof(item)) === 'object' && item.constructor ? item.constructor.name || '[Unknown Object]' : typeof item === 'undefined' ? 'undefined' : _typeof(item)));
        }

        throw path.buildCodeFrameError('Preconditions cannot have side effects.');
      },
      ExpressionStatement: function ExpressionStatement(statement) {
        if (!NodePath(statement)) {
          throw new TypeError('Value of argument "statement" violates contract, expected NodePath got ' + (statement === null ? 'null' : (typeof statement === 'undefined' ? 'undefined' : _typeof(statement)) === 'object' && statement.constructor ? statement.constructor.name || '[Unknown Object]' : typeof statement === 'undefined' ? 'undefined' : _typeof(statement)));
        }

        var condition = statement.get('expression');

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
        }

        var message = undefined;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }

        if (condition.isSequenceExpression()) {
          var expressions = condition.get('expressions');
          condition = expressions[0];

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
          }

          message = expressions[1].node;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }
        } else {
          message = t.stringLiteral('Function ' + name + 'precondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }
        }
        statement.replaceWith(guard({
          condition: staticCheck(condition),
          message: message
        }));
      }
    });

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function assemblePostcondition(path) {
    function _ref3(_id5) {
      if (!Identifier(_id5)) {
        throw new TypeError('Function "assemblePostcondition" return value violates contract, expected Identifier got ' + (_id5 === null ? 'null' : (typeof _id5 === 'undefined' ? 'undefined' : _typeof(_id5)) === 'object' && _id5.constructor ? _id5.constructor.name || '[Unknown Object]' : typeof _id5 === 'undefined' ? 'undefined' : _typeof(_id5)));
      }

      return _id5;
    }

    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract, expected NodePath got ' + (body === null ? 'null' : (typeof body === 'undefined' ? 'undefined' : _typeof(body)) === 'object' && body.constructor ? body.constructor.name || '[Unknown Object]' : typeof body === 'undefined' ? 'undefined' : _typeof(body)));
    }

    var fn = path.getFunctionParent();

    if (!NodePath(fn)) {
      throw new TypeError('Value of variable "fn" violates contract, expected NodePath got ' + (fn === null ? 'null' : (typeof fn === 'undefined' ? 'undefined' : _typeof(fn)) === 'object' && fn.constructor ? fn.constructor.name || '[Unknown Object]' : typeof fn === 'undefined' ? 'undefined' : _typeof(fn)));
    }

    var name = fn.node.id ? '"' + fn.node.id.name + '" ' : ' ';
    var conditions = [];

    if (!(Array.isArray(conditions) && conditions.every(function (item) {
      return Node(item);
    }))) {
      throw new TypeError('Value of variable "conditions" violates contract, expected Node[] got ' + (conditions === null ? 'null' : (typeof conditions === 'undefined' ? 'undefined' : _typeof(conditions)) === 'object' && conditions.constructor ? conditions.constructor.name || '[Unknown Object]' : typeof conditions === 'undefined' ? 'undefined' : _typeof(conditions)));
    }

    var captures = [];

    if (!(Array.isArray(captures) && captures.every(function (item) {
      return Node(item);
    }))) {
      throw new TypeError('Value of variable "captures" violates contract, expected Node[] got ' + (captures === null ? 'null' : (typeof captures === 'undefined' ? 'undefined' : _typeof(captures)) === 'object' && captures.constructor ? captures.constructor.name || '[Unknown Object]' : typeof captures === 'undefined' ? 'undefined' : _typeof(captures)));
    }

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
      }

      var message = undefined;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      } else {
        message = t.stringLiteral('Function ' + name + 'postcondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      }
      conditions.push(guard({
        condition: staticCheck(condition),
        message: message
      }));
    } else {
      body.traverse({
        "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
          if (!NodePath(item)) {
            throw new TypeError('Value of argument "item" violates contract, expected NodePath got ' + (item === null ? 'null' : (typeof item === 'undefined' ? 'undefined' : _typeof(item)) === 'object' && item.constructor ? item.constructor.name || '[Unknown Object]' : typeof item === 'undefined' ? 'undefined' : _typeof(item)));
          }

          throw path.buildCodeFrameError('Postconditions cannot have side effects.');
        },
        CallExpression: function CallExpression(call) {
          if (!NodePath(call)) {
            throw new TypeError('Value of argument "call" violates contract, expected NodePath got ' + (call === null ? 'null' : (typeof call === 'undefined' ? 'undefined' : _typeof(call)) === 'object' && call.constructor ? call.constructor.name || '[Unknown Object]' : typeof call === 'undefined' ? 'undefined' : _typeof(call)));
          }

          var callee = call.get('callee');

          if (!NodePath(callee)) {
            throw new TypeError('Value of variable "callee" violates contract, expected NodePath got ' + (callee === null ? 'null' : (typeof callee === 'undefined' ? 'undefined' : _typeof(callee)) === 'object' && callee.constructor ? callee.constructor.name || '[Unknown Object]' : typeof callee === 'undefined' ? 'undefined' : _typeof(callee)));
          }

          var args = call.get('arguments');

          if (!(Array.isArray(args) && args.every(function (item) {
            return NodePath(item);
          }))) {
            throw new TypeError('Value of variable "args" violates contract, expected NodePath[] got ' + (args === null ? 'null' : (typeof args === 'undefined' ? 'undefined' : _typeof(args)) === 'object' && args.constructor ? args.constructor.name || '[Unknown Object]' : typeof args === 'undefined' ? 'undefined' : _typeof(args)));
          }

          if (!callee.isIdentifier() || callee.node.name !== OLD_VALUE_NAME || call.scope.hasBinding(OLD_VALUE_NAME) || args.length === 0) {
            return;
          }
          var argument = args[0];

          if (!NodePath(argument)) {
            throw new TypeError('Value of variable "argument" violates contract, expected NodePath got ' + (argument === null ? 'null' : (typeof argument === 'undefined' ? 'undefined' : _typeof(argument)) === 'object' && argument.constructor ? argument.constructor.name || '[Unknown Object]' : typeof argument === 'undefined' ? 'undefined' : _typeof(argument)));
          }

          var id = call.scope.generateUidIdentifierBasedOnNode(argument.node);
          fn.scope.push({ id: id, init: argument.node, kind: 'const' });
          call.replaceWith(id);
        },
        ExpressionStatement: function ExpressionStatement(statement) {
          if (!NodePath(statement)) {
            throw new TypeError('Value of argument "statement" violates contract, expected NodePath got ' + (statement === null ? 'null' : (typeof statement === 'undefined' ? 'undefined' : _typeof(statement)) === 'object' && statement.constructor ? statement.constructor.name || '[Unknown Object]' : typeof statement === 'undefined' ? 'undefined' : _typeof(statement)));
          }

          var condition = statement.get('expression');

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
          }

          var message = undefined;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }

          if (condition.isSequenceExpression()) {
            var expressions = condition.get('expressions');
            condition = expressions[0];

            if (!NodePath(condition)) {
              throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
            }

            message = expressions[1].node;

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
            }
          } else {
            message = t.stringLiteral('Function ' + name + 'postcondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
            }
          }
          statement.replaceWith(guard({
            condition: staticCheck(condition),
            message: message
          }));
        }
      });
      conditions.push.apply(conditions, _toConsumableArray(body.node.body));
    }

    var id = path.scope.generateUidIdentifier((fn.node.id ? fn.node.id.name : 'check') + 'Postcondition');

    fn.get('body').get('body')[0].insertBefore(guardFn({
      id: id,
      conditions: conditions,
      it: returnId
    }));

    path.remove();
    return _ref3(id);
  }

  function assembleAssertion(path) {
    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract, expected NodePath got ' + (body === null ? 'null' : (typeof body === 'undefined' ? 'undefined' : _typeof(body)) === 'object' && body.constructor ? body.constructor.name || '[Unknown Object]' : typeof body === 'undefined' ? 'undefined' : _typeof(body)));
    }

    var fn = path.getFunctionParent();

    if (!(fn == null || NodePath(fn))) {
      throw new TypeError('Value of variable "fn" violates contract, expected ?NodePath got ' + (fn === null ? 'null' : (typeof fn === 'undefined' ? 'undefined' : _typeof(fn)) === 'object' && fn.constructor ? fn.constructor.name || '[Unknown Object]' : typeof fn === 'undefined' ? 'undefined' : _typeof(fn)));
    }

    var name = fn && fn.node && fn.node.id ? '"' + fn.node.id.name + '"' : '';
    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
      }

      var message = undefined;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      } else if (name) {
        message = t.stringLiteral('Function ' + name + ' assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      } else {
        message = t.stringLiteral('Assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      }
      path.replaceWith(guard({
        condition: staticCheck(condition),
        message: message
      }));
      return;
    }

    body.traverse({
      "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
        if (!NodePath(item)) {
          throw new TypeError('Value of argument "item" violates contract, expected NodePath got ' + (item === null ? 'null' : (typeof item === 'undefined' ? 'undefined' : _typeof(item)) === 'object' && item.constructor ? item.constructor.name || '[Unknown Object]' : typeof item === 'undefined' ? 'undefined' : _typeof(item)));
        }

        throw path.buildCodeFrameError('Assertions cannot have side effects.');
      },
      ExpressionStatement: function ExpressionStatement(statement) {
        if (!NodePath(statement)) {
          throw new TypeError('Value of argument "statement" violates contract, expected NodePath got ' + (statement === null ? 'null' : (typeof statement === 'undefined' ? 'undefined' : _typeof(statement)) === 'object' && statement.constructor ? statement.constructor.name || '[Unknown Object]' : typeof statement === 'undefined' ? 'undefined' : _typeof(statement)));
        }

        var condition = statement.get('expression');

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
        }

        var message = undefined;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }

        if (condition.isSequenceExpression()) {
          var expressions = condition.get('expressions');
          condition = expressions[0];

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
          }

          message = expressions[1].node;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }
        } else if (name) {
          message = t.stringLiteral('Function ' + name + ' assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }
        } else {
          message = t.stringLiteral('Assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }
        }
        statement.replaceWith(guard({
          condition: staticCheck(condition),
          message: message
        }));
      }
    });

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function assembleInvariant(path) {
    function _ref5(_id12) {
      if (!Identifier(_id12)) {
        throw new TypeError('Function "assembleInvariant" return value violates contract, expected Identifier got ' + (_id12 === null ? 'null' : (typeof _id12 === 'undefined' ? 'undefined' : _typeof(_id12)) === 'object' && _id12.constructor ? _id12.constructor.name || '[Unknown Object]' : typeof _id12 === 'undefined' ? 'undefined' : _typeof(_id12)));
      }

      return _id12;
    }

    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract, expected NodePath got ' + (body === null ? 'null' : (typeof body === 'undefined' ? 'undefined' : _typeof(body)) === 'object' && body.constructor ? body.constructor.name || '[Unknown Object]' : typeof body === 'undefined' ? 'undefined' : _typeof(body)));
    }

    var fn = path.getFunctionParent();

    if (!NodePath(fn)) {
      throw new TypeError('Value of variable "fn" violates contract, expected NodePath got ' + (fn === null ? 'null' : (typeof fn === 'undefined' ? 'undefined' : _typeof(fn)) === 'object' && fn.constructor ? fn.constructor.name || '[Unknown Object]' : typeof fn === 'undefined' ? 'undefined' : _typeof(fn)));
    }

    var name = fn.node.id ? '"' + fn.node.id.name + '" ' : ' ';
    var conditions = [];

    if (!(Array.isArray(conditions) && conditions.every(function (item) {
      return Node(item);
    }))) {
      throw new TypeError('Value of variable "conditions" violates contract, expected Node[] got ' + (conditions === null ? 'null' : (typeof conditions === 'undefined' ? 'undefined' : _typeof(conditions)) === 'object' && conditions.constructor ? conditions.constructor.name || '[Unknown Object]' : typeof conditions === 'undefined' ? 'undefined' : _typeof(conditions)));
    }

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
      }

      var message = undefined;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      } else {
        message = t.stringLiteral('Function ' + name + 'invariant failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
        }
      }
      conditions.push(guard({
        condition: staticCheck(condition),
        message: message
      }));
    } else {
      body.traverse({
        "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
          if (!NodePath(item)) {
            throw new TypeError('Value of argument "item" violates contract, expected NodePath got ' + (item === null ? 'null' : (typeof item === 'undefined' ? 'undefined' : _typeof(item)) === 'object' && item.constructor ? item.constructor.name || '[Unknown Object]' : typeof item === 'undefined' ? 'undefined' : _typeof(item)));
          }

          throw path.buildCodeFrameError('Invariants cannot have side effects.');
        },
        ExpressionStatement: function ExpressionStatement(statement) {
          if (!NodePath(statement)) {
            throw new TypeError('Value of argument "statement" violates contract, expected NodePath got ' + (statement === null ? 'null' : (typeof statement === 'undefined' ? 'undefined' : _typeof(statement)) === 'object' && statement.constructor ? statement.constructor.name || '[Unknown Object]' : typeof statement === 'undefined' ? 'undefined' : _typeof(statement)));
          }

          var condition = statement.get('expression');

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
          }

          var message = undefined;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
          }

          if (condition.isSequenceExpression()) {
            var expressions = condition.get('expressions');
            condition = expressions[0];

            if (!NodePath(condition)) {
              throw new TypeError('Value of variable "condition" violates contract, expected NodePath got ' + (condition === null ? 'null' : (typeof condition === 'undefined' ? 'undefined' : _typeof(condition)) === 'object' && condition.constructor ? condition.constructor.name || '[Unknown Object]' : typeof condition === 'undefined' ? 'undefined' : _typeof(condition)));
            }

            message = expressions[1].node;

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
            }
          } else {
            message = t.stringLiteral('Function ' + name + 'invariant failed: ' + (0, _babelGenerator2.default)(condition.node).code);

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract, expected ?Node got ' + (message === null ? 'null' : (typeof message === 'undefined' ? 'undefined' : _typeof(message)) === 'object' && message.constructor ? message.constructor.name || '[Unknown Object]' : typeof message === 'undefined' ? 'undefined' : _typeof(message)));
            }
          }
          statement.replaceWith(guard({
            condition: staticCheck(condition),
            message: message
          }));
        }
      });
      conditions.push.apply(conditions, _toConsumableArray(body.node.body));
    }

    var id = path.scope.generateUidIdentifier((fn.node.id ? fn.node.id.name : 'check') + 'Invariant');
    path.parentPath.get('body')[0].insertBefore(guardFn({
      id: id,
      conditions: conditions,
      it: returnId
    }));
    path.remove();
    return _ref5(id);
  }

  function expression(input) {
    function _ref6(_id15) {
      if (!(typeof _id15 === 'function')) {
        throw new TypeError('Function "expression" return value violates contract, expected Function got ' + (_id15 === null ? 'null' : (typeof _id15 === 'undefined' ? 'undefined' : _typeof(_id15)) === 'object' && _id15.constructor ? _id15.constructor.name || '[Unknown Object]' : typeof _id15 === 'undefined' ? 'undefined' : _typeof(_id15)));
      }

      return _id15;
    }

    if (!(typeof input === 'string')) {
      throw new TypeError('Value of argument "input" violates contract, expected string got ' + (input === null ? 'null' : (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && input.constructor ? input.constructor.name || '[Unknown Object]' : typeof input === 'undefined' ? 'undefined' : _typeof(input)));
    }

    var fn = template(input);

    if (!(typeof fn === 'function')) {
      throw new TypeError('Value of variable "fn" violates contract, expected Function got ' + (fn === null ? 'null' : (typeof fn === 'undefined' ? 'undefined' : _typeof(fn)) === 'object' && fn.constructor ? fn.constructor.name || '[Unknown Object]' : typeof fn === 'undefined' ? 'undefined' : _typeof(fn)));
    }

    return _ref6(function () {
      var node = fn.apply(undefined, arguments);

      if (!Node(node)) {
        throw new TypeError('Value of variable "node" violates contract, expected Node got ' + (node === null ? 'null' : (typeof node === 'undefined' ? 'undefined' : _typeof(node)) === 'object' && node.constructor ? node.constructor.name || '[Unknown Object]' : typeof node === 'undefined' ? 'undefined' : _typeof(node)));
      }

      return getExpression(node);
    });
  }

  function staticCheck(expression) {
    function _ref7(_id16) {
      if (!NodePath(_id16)) {
        throw new TypeError('Function "staticCheck" return value violates contract, expected NodePath got ' + (_id16 === null ? 'null' : (typeof _id16 === 'undefined' ? 'undefined' : _typeof(_id16)) === 'object' && _id16.constructor ? _id16.constructor.name || '[Unknown Object]' : typeof _id16 === 'undefined' ? 'undefined' : _typeof(_id16)));
      }

      return _id16;
    }

    if (!NodePath(expression)) {
      throw new TypeError('Value of argument "expression" violates contract, expected NodePath got ' + (expression === null ? 'null' : (typeof expression === 'undefined' ? 'undefined' : _typeof(expression)) === 'object' && expression.constructor ? expression.constructor.name || '[Unknown Object]' : typeof expression === 'undefined' ? 'undefined' : _typeof(expression)));
    }

    var _expression$evaluate = expression.evaluate();

    var confident = _expression$evaluate.confident;
    var value = _expression$evaluate.value;

    if (confident && !value) {
      throw expression.buildCodeFrameError('Contract always fails.');
    }

    return _ref7(expression);
  }

  return _ref({
    visitor: {
      Function: function Function(fn, _ref9) {
        var opts = _ref9.opts;

        if (!NodePath(fn)) {
          throw new TypeError('Value of argument "fn" violates contract, expected NodePath got ' + (fn === null ? 'null' : (typeof fn === 'undefined' ? 'undefined' : _typeof(fn)) === 'object' && fn.constructor ? fn.constructor.name || '[Unknown Object]' : typeof fn === 'undefined' ? 'undefined' : _typeof(fn)));
        }

        if (fn.isArrowFunctionExpression() && !fn.get('body').isBlockStatement()) {
          // Naked arrow functions cannot contain contracts.
          return;
        }
        fn.traverse({
          Function: function Function(path) {
            if (!NodePath(path)) {
              throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
            }

            // This will be handled by the outer visitor, so skip it.
            path.skip();
          },
          LabeledStatement: function LabeledStatement(path) {
            if (!NodePath(path)) {
              throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
            }

            var label = path.get('label');

            if (!NodePath(label)) {
              throw new TypeError('Value of variable "label" violates contract, expected NodePath got ' + (label === null ? 'null' : (typeof label === 'undefined' ? 'undefined' : _typeof(label)) === 'object' && label.constructor ? label.constructor.name || '[Unknown Object]' : typeof label === 'undefined' ? 'undefined' : _typeof(label)));
            }

            if (opts.strip || opts.env && opts.env[process.env.NODE_ENV] && opts.env[process.env.NODE_ENV].strip) {
              if (label.node.name === PRECONDITION_NAME || label.node.name === POSTCONDITION_NAME || label.node.name === INVARIANT_NAME || label.node.name === ASSERT_NAME) {
                path.remove();
              }
              return;
            }

            var id = undefined;

            if (!(id == null || Identifier(id))) {
              throw new TypeError('Value of variable "id" violates contract, expected ?Identifier got ' + (id === null ? 'null' : (typeof id === 'undefined' ? 'undefined' : _typeof(id)) === 'object' && id.constructor ? id.constructor.name || '[Unknown Object]' : typeof id === 'undefined' ? 'undefined' : _typeof(id)));
            }

            var children = undefined;

            if (!(children == null || Array.isArray(children) && children.every(function (item) {
              return NodePath(item);
            }))) {
              throw new TypeError('Value of variable "children" violates contract, expected ?NodePath[] got ' + (children === null ? 'null' : (typeof children === 'undefined' ? 'undefined' : _typeof(children)) === 'object' && children.constructor ? children.constructor.name || '[Unknown Object]' : typeof children === 'undefined' ? 'undefined' : _typeof(children)));
            }

            var parent = fn;

            if (!NodePath(parent)) {
              throw new TypeError('Value of variable "parent" violates contract, expected NodePath got ' + (parent === null ? 'null' : (typeof parent === 'undefined' ? 'undefined' : _typeof(parent)) === 'object' && parent.constructor ? parent.constructor.name || '[Unknown Object]' : typeof parent === 'undefined' ? 'undefined' : _typeof(parent)));
            }

            if (label.node.name === PRECONDITION_NAME) {
              assemblePrecondition(path);
              return;
            } else if (label.node.name === POSTCONDITION_NAME) {
              id = assemblePostcondition(path);
              children = fn.get('body').get('body');

              if (!(children == null || Array.isArray(children) && children.every(function (item) {
                return NodePath(item);
              }))) {
                throw new TypeError('Value of variable "children" violates contract, expected ?NodePath[] got ' + (children === null ? 'null' : (typeof children === 'undefined' ? 'undefined' : _typeof(children)) === 'object' && children.constructor ? children.constructor.name || '[Unknown Object]' : typeof children === 'undefined' ? 'undefined' : _typeof(children)));
              }
            } else if (label.node.name === ASSERT_NAME) {
              assembleAssertion(path);
              return;
            } else if (label.node.name === INVARIANT_NAME) {
              id = assembleInvariant(path);
              parent = path.findParent(t.isBlockStatement);

              if (!NodePath(parent)) {
                throw new TypeError('Value of variable "parent" violates contract, expected NodePath got ' + (parent === null ? 'null' : (typeof parent === 'undefined' ? 'undefined' : _typeof(parent)) === 'object' && parent.constructor ? parent.constructor.name || '[Unknown Object]' : typeof parent === 'undefined' ? 'undefined' : _typeof(parent)));
              }

              children = parent.get('body');

              if (!(children == null || Array.isArray(children) && children.every(function (item) {
                return NodePath(item);
              }))) {
                throw new TypeError('Value of variable "children" violates contract, expected ?NodePath[] got ' + (children === null ? 'null' : (typeof children === 'undefined' ? 'undefined' : _typeof(children)) === 'object' && children.constructor ? children.constructor.name || '[Unknown Object]' : typeof children === 'undefined' ? 'undefined' : _typeof(children)));
              }

              var first = children[0];

              if (!NodePath(first)) {
                throw new TypeError('Value of variable "first" violates contract, expected NodePath got ' + (first === null ? 'null' : (typeof first === 'undefined' ? 'undefined' : _typeof(first)) === 'object' && first.constructor ? first.constructor.name || '[Unknown Object]' : typeof first === 'undefined' ? 'undefined' : _typeof(first)));
              }

              first.insertAfter(t.expressionStatement(t.callExpression(id, [])));
            }
            parent.traverse({
              Function: function Function(path) {
                if (!NodePath(path)) {
                  throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
                }

                // This will be handled by the outer visitor, so skip it.
                path.skip();
              },
              ReturnStatement: function ReturnStatement(statement) {
                if (!NodePath(statement)) {
                  throw new TypeError('Value of argument "statement" violates contract, expected NodePath got ' + (statement === null ? 'null' : (typeof statement === 'undefined' ? 'undefined' : _typeof(statement)) === 'object' && statement.constructor ? statement.constructor.name || '[Unknown Object]' : typeof statement === 'undefined' ? 'undefined' : _typeof(statement)));
                }

                statement.get('argument').replaceWith(t.callExpression(id, [statement.node.argument]));
              }
            });
            var last = children[children.length - 1];

            if (!NodePath(last)) {
              throw new TypeError('Value of variable "last" violates contract, expected NodePath got ' + (last === null ? 'null' : (typeof last === 'undefined' ? 'undefined' : _typeof(last)) === 'object' && last.constructor ? last.constructor.name || '[Unknown Object]' : typeof last === 'undefined' ? 'undefined' : _typeof(last)));
            }

            if (!last.isReturnStatement()) {
              last.insertAfter(t.expressionStatement(t.callExpression(id, [])));
            }
          }
        });
      },
      LabeledStatement: function LabeledStatement(path, _ref10) {
        var opts = _ref10.opts;

        if (!NodePath(path)) {
          throw new TypeError('Value of argument "path" violates contract, expected NodePath got ' + (path === null ? 'null' : (typeof path === 'undefined' ? 'undefined' : _typeof(path)) === 'object' && path.constructor ? path.constructor.name || '[Unknown Object]' : typeof path === 'undefined' ? 'undefined' : _typeof(path)));
        }

        var label = path.get('label');

        if (!NodePath(label)) {
          throw new TypeError('Value of variable "label" violates contract, expected NodePath got ' + (label === null ? 'null' : (typeof label === 'undefined' ? 'undefined' : _typeof(label)) === 'object' && label.constructor ? label.constructor.name || '[Unknown Object]' : typeof label === 'undefined' ? 'undefined' : _typeof(label)));
        }

        if (label.node.name === ASSERT_NAME) {
          if (opts.strip || opts.env && opts.env[process.env.NODE_ENV] && opts.env[process.env.NODE_ENV].strip) {
            path.remove();
          } else {
            assembleAssertion(path);
          }
          return;
        }
      }
    }
  });
};

var _babelGenerator = require('babel-generator');

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _typeof(obj) { return obj && typeof Symbol !== "undefined" && obj.constructor === Symbol ? "symbol" : typeof obj; }

var Plugin = function Plugin(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && Visitors(input.visitor);
};

var PluginParams = function PluginParams(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && input.types instanceof Object && typeof input.template === 'function';
};

var Visitors = function Visitors(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object';
};

var Visitor = function Visitor(input) {
  return typeof input === 'function';
};

var Node = function Node(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && typeof input.type === 'string' && (input.node === undefined || input.node == null);
};

var Literal = function Literal(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && (input.type === 'StringLiteral' || input.type === 'BooleanLiteral' || input.type === 'NumericLiteral' || input.type === 'NullLiteral' || input.type === 'RegExpLiteral');
};

var Identifier = function Identifier(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && typeof input.type === 'string' && typeof input.name === 'string';
};

var Scope = function Scope(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object';
};

var NodePath = function NodePath(input) {
  return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && typeof input.type === 'string' && Node(input.node) && Scope(input.scope);
};

/**
 * # Design By Contract Transformer
 */